import { Boolean, Context, Effect, Layer, MutableRef, Option, Readable, Record, Stream, Subscribable, } from 'effect';
import { dual, identity } from 'effect/Function';
import { hasProperty } from 'effect/Predicate';
import * as Computed from './Computed.js';
import { deepEquals, EffectBase, getExitEquivalence, getOrInitializeCore, getSetDelete, interruptCore, interruptCoreAndWait, makeCore, makeDeferredRef, sendEvent, streamExit, } from './internal.js';
import * as Provide from './Provide.js';
import * as Versioned from './Versioned.js';
export const TypeId = Symbol.for('@typed/LazyRef');
function tupleGetSetDelete(gsds) {
    return {
        get: Effect.all(gsds.map((gsd) => gsd.get), { concurrency: 'unbounded' }),
        set: (values) => Effect.all(gsds.map((gsd, i) => gsd.set(values[i])), { concurrency: 'unbounded' }),
        delete: Effect.map(Effect.all(gsds.map((gsd) => gsd.delete), { concurrency: 'unbounded' }), Option.all),
        version: Effect.all(gsds.map((gsd) => gsd.version), { concurrency: 'unbounded' }).pipe(Effect.map((versions) => versions.reduce(sum, 0))),
    };
}
function structGetSetDelete(gsds) {
    return {
        get: Effect.all(Record.map(gsds, (gsd) => gsd.get), { concurrency: 'unbounded' }),
        set: (values) => Effect.all(Record.map(gsds, (gsd, i) => gsd.set(values[i])), { concurrency: 'unbounded' }),
        delete: Effect.map(Effect.all(Record.map(gsds, (gsd) => gsd.delete), { concurrency: 'unbounded' }), Option.all),
        version: Effect.all(Object.values(gsds).map((gsd) => gsd.version), { concurrency: 'unbounded' }).pipe(Effect.map((versions) => versions.reduce(sum, 0))),
    };
}
const variance = {
    _A: (_) => _,
    _E: (_) => _,
    _R: (_) => _,
};
class VariancesImpl extends EffectBase {
    [TypeId] = variance;
    [Computed.ComputedTypeId] = variance;
    [Subscribable.TypeId] = Subscribable.TypeId;
    [Readable.TypeId] = Readable.TypeId;
}
class SubscriptionRefImpl extends VariancesImpl {
    core;
    version;
    shutdown;
    awaitShutdown;
    subscriberCount;
    getSetDelete;
    get;
    changes;
    constructor(core) {
        super();
        this.core = core;
        this.version = Effect.sync(() => core.deferredRef.version);
        this.shutdown = Effect.provide(interruptCore(core), core.runtime.context);
        this.awaitShutdown = Effect.provide(interruptCoreAndWait(core), core.runtime.context);
        this.subscriberCount = Effect.sync(() => MutableRef.get(core.pubsub.subscriberCount));
        this.get = this.toEffect();
        this.getSetDelete = getSetDelete(core);
        this.changes = Stream.flatten(Stream.fromPubSub(core.pubsub));
    }
    toEffect() {
        return getOrInitializeCore(this.core, true);
    }
    runUpdates(f) {
        return this.core.semaphore.withPermits(1)(f(this.getSetDelete));
    }
}
export const fromEffect = (effect, options) => Effect.map(makeCore(effect, options), (core) => new SubscriptionRefImpl(core));
export const fromStream = (stream, options) => makeDeferredRef(getExitEquivalence(options?.eq ?? deepEquals)).pipe(Effect.bindTo('deferredRef'), Effect.bind('core', ({ deferredRef }) => makeCore(deferredRef, options)), Effect.tap(({ core, deferredRef }) => Effect.forkIn(stream.pipe(streamExit, Stream.runForEach((exit) => deferredRef.done(exit) ? sendEvent(core, exit) : Effect.void)), core.scope)), Effect.map(({ core }) => new SubscriptionRefImpl(core)));
export function make(input, options) {
    return Effect.isEffect(input) ? fromEffect(input, options) : fromStream(input, options);
}
export function of(value, options) {
    return make(Effect.succeed(value), options);
}
export function failCause(cause, options) {
    return make(Effect.failCause(cause), options);
}
export function fail(error, options) {
    return make(Effect.fail(error), options);
}
export function get(ref) {
    return ref.get;
}
export const set = dual(2, function set(ref, a) {
    return ref.runUpdates(({ set }) => set(a));
});
export function reset(ref) {
    return ref.runUpdates((ref) => ref.delete);
}
export { reset as delete };
export const modify = dual(2, function modify(ref, f) {
    return ref.runUpdates(({ set, get }) => get.pipe(Effect.map(f), Effect.flatMap(([b, a]) => set(a).pipe(Effect.as(b)))));
});
export const modifyEffect = dual(2, function modifyEffect(ref, f) {
    return ref.runUpdates(({ set, get }) => get.pipe(Effect.flatMap(f), Effect.flatMap(([b, a]) => set(a).pipe(Effect.as(b)))));
});
export const update = dual(2, function update(ref, f) {
    return modify(ref, (a) => {
        const a2 = f(a);
        return [a2, a2];
    });
});
export const getAndUpdate = dual(2, function getAndUpdate(ref, f) {
    return modify(ref, (a) => {
        const a2 = f(a);
        return [a, a2];
    });
});
export const updateEffect = dual(2, function updateEffect(ref, f) {
    return modifyEffect(ref, (a) => Effect.map(f(a), (a2) => [a2, a2]));
});
export const getAndUpdateEffect = dual(2, function getAndUpdateEffect(ref, f) {
    return modifyEffect(ref, (a) => Effect.map(f(a), (a2) => [a, a2]));
});
export const getAndSet = dual(2, function getAndSet(ref, a) {
    return getAndUpdate(ref, () => a);
});
export const mapEffect = dual((args) => Computed.isComputed(args[0]), function mapEffect(ref, f, options) {
    return Computed.makeComputed(Effect.flatMap(ref, f), Stream.flatMap(ref.changes, f, options), ref.version);
});
export const map = dual(2, function map(ref, f) {
    return Computed.makeComputed(Effect.map(ref, f), Stream.map(ref.changes, f), ref.version);
});
export const proxy = (source) => {
    const target = {};
    return new Proxy(target, {
        get(self, prop) {
            return (self[prop] ??= map(source, (a) => a[prop]));
        },
    });
};
export function isSubscriptionRef(value) {
    return hasProperty(value, TypeId);
}
const sum = (a, b) => a + b;
class TupleImpl extends VariancesImpl {
    refs;
    [TypeId] = variance;
    [Computed.ComputedTypeId] = variance;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    constructor(refs) {
        super();
        this.refs = refs;
        this.subscriberCount = Effect.all(refs.map((ref) => ref.subscriberCount), { concurrency: 'unbounded' }).pipe(Effect.map((counts) => counts.reduce(sum, 0)));
        this.shutdown = Effect.all(refs.map((ref) => ref.shutdown), { concurrency: 'unbounded' });
        this.awaitShutdown = Effect.all(refs.map((ref) => ref.awaitShutdown), { concurrency: 'unbounded' });
        this.version = Effect.all(refs.map((ref) => ref.version), { concurrency: 'unbounded' }).pipe(Effect.map((versions) => versions.reduce(sum, 0)));
        this.changes = Stream.zipLatestAll(...refs.map((ref) => ref.changes));
    }
    get get() {
        return this;
    }
    toEffect() {
        return Effect.all(this.refs, { concurrency: 'unbounded' });
    }
    runUpdates(f) {
        return tupleRunUpdates(this.refs, (gsds) => f(tupleGetSetDelete(gsds)));
    }
}
export function tuple(...refs) {
    const hasNonSubscriptionRefs = refs.some((ref) => !isSubscriptionRef(ref));
    if (hasNonSubscriptionRefs) {
        return Computed.fromVersioned(Versioned.tuple(...refs));
    }
    return new TupleImpl(refs);
}
export function struct(refs) {
    const hasNonSubscriptionRefs = Object.values(refs).some((ref) => !isSubscriptionRef(ref));
    if (hasNonSubscriptionRefs) {
        return Computed.fromVersioned(Versioned.struct(refs));
    }
    return new StructImpl(refs);
}
class StructImpl extends VariancesImpl {
    refs;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    constructor(refs) {
        super();
        this.refs = refs;
        const refValues = Object.values(refs);
        this.shutdown = Effect.all(refValues.map((ref) => ref.shutdown), { concurrency: 'unbounded' });
        this.awaitShutdown = Effect.all(refValues.map((ref) => ref.awaitShutdown), { concurrency: 'unbounded' });
        this.subscriberCount = Effect.all(refValues.map((ref) => ref.subscriberCount), { concurrency: 'unbounded' }).pipe(Effect.map((counts) => counts.reduce(sum, 0)));
        this.version = Effect.all(refValues.map((ref) => ref.version), { concurrency: 'unbounded' }).pipe(Effect.map((versions) => versions.reduce(sum, 0)));
        this.changes = Stream.zipLatestAll(...refValues.map((ref) => ref.changes));
    }
    get get() {
        return this;
    }
    toEffect() {
        return Effect.all(this.refs, { concurrency: 'unbounded' });
    }
    runUpdates(f) {
        return structRunUpdates(this.refs, (gsds) => f(structGetSetDelete(gsds)));
    }
}
function tupleRunUpdates(refs, accumulated) {
    if (refs.length === 0) {
        return accumulated([]);
    }
    const [first, ...rest] = refs;
    return first.runUpdates((firstGsd) => tupleRunUpdates(rest, (restGsds) => accumulated([firstGsd, ...restGsds])));
}
function structRunUpdates(refs, accumulated) {
    const entries = Object.entries(refs);
    return tupleRunUpdates(entries.map(([_, ref]) => ref), (gsds) => accumulated(Object.fromEntries(entries.map(([key], i) => [key, gsds[i]]))));
}
class TaggedImpl extends VariancesImpl {
    id;
    tag;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    runUpdates;
    constructor(id) {
        super();
        this.id = id;
        this.id = id;
        this.tag = Context.GenericTag(id);
        this.shutdown = Effect.flatMap(this.tag, (ref) => ref.shutdown);
        this.awaitShutdown = Effect.flatMap(this.tag, (ref) => ref.awaitShutdown);
        this.subscriberCount = Effect.flatMap(this.tag, (ref) => ref.subscriberCount);
        this.version = Effect.flatMap(this.tag, (ref) => ref.version);
        this.changes = Stream.unwrap(Effect.map(this.tag, (ref) => ref.changes));
        this.runUpdates = (f) => Effect.flatMap(this.tag, (ref) => ref.runUpdates(f));
    }
    get get() {
        return this;
    }
    toEffect() {
        return Effect.flatten(this.tag);
    }
    of(value) {
        return this.make(Effect.succeed(value));
    }
    make(fxOrEffect, options) {
        return Layer.scoped(this.tag, make(fxOrEffect, options).pipe(Effect.map((x) => (options?.drop ? drop(x, options.drop) : x)), Effect.map((x) => (options?.take ? take(x, options.take) : x))));
    }
    layer(make) {
        return Layer.scoped(this.tag, make);
    }
}
export function Tag(id) {
    return () => {
        const base = new TaggedImpl(id);
        function klass() {
            return base;
        }
        Object.setPrototypeOf(klass, base);
        return klass;
    };
}
class FromTag extends VariancesImpl {
    tag;
    f;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    runUpdates;
    constructor(tag, f) {
        super();
        this.tag = tag;
        this.f = f;
        this.shutdown = Effect.flatMap(this.tag, (s) => this.f(s).shutdown);
        this.awaitShutdown = Effect.flatMap(this.tag, (s) => this.f(s).awaitShutdown);
        this.subscriberCount = Effect.flatMap(this.tag, (s) => this.f(s).subscriberCount);
        this.version = Effect.flatMap(this.tag, (s) => this.f(s).version);
        this.changes = Stream.unwrap(Effect.map(this.tag, (s) => this.f(s).changes));
        this.runUpdates = (f) => Effect.flatMap(this.tag, (s) => this.f(s).runUpdates(f));
    }
    toEffect() {
        return Effect.flatMap(this.tag, this.f);
    }
    get get() {
        return this;
    }
}
export function fromTag(tag, f) {
    return new FromTag(tag, f);
}
export const when = dual(2, function when(ref, matchers) {
    return Computed.makeComputed(Effect.map(ref, Boolean.match(matchers)), Stream.map(ref.changes, Boolean.match(matchers)), ref.version);
});
class SubscriptionRefProvide extends VariancesImpl {
    ref;
    provide;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    runUpdates;
    constructor(ref, provide) {
        super();
        this.ref = ref;
        this.provide = provide;
        this.shutdown = Provide.provideToEffect(ref.shutdown, provide);
        this.awaitShutdown = Provide.provideToEffect(ref.awaitShutdown, provide);
        this.subscriberCount = Provide.provideToEffect(ref.subscriberCount, provide);
        this.version = Provide.provideToEffect(ref.version, provide);
        this.changes = Provide.provideToStream(ref.changes, provide);
        this.runUpdates = (f) => Provide.provideToEffect(ref.runUpdates((gsd) => f(gsd)), provide);
    }
    toEffect() {
        return Provide.provideToEffect(this.ref, this.provide);
    }
    get get() {
        return this;
    }
}
function provide_(ref, provide) {
    if (ref instanceof SubscriptionRefProvide) {
        return new SubscriptionRefProvide(ref.ref, Provide.merge(ref.provide, provide));
    }
    return new SubscriptionRefProvide(ref, provide);
}
export const provide = dual(2, function provide(ref, contextOrLayerOrRuntime) {
    const provide = Context.isContext(contextOrLayerOrRuntime)
        ? Provide.ProvideContext(contextOrLayerOrRuntime)
        : Layer.isLayer(contextOrLayerOrRuntime)
            ? Provide.ProvideLayer(contextOrLayerOrRuntime)
            : Provide.ProvideRuntime(contextOrLayerOrRuntime);
    if (isSubscriptionRef(ref)) {
        return provide_(ref, provide);
    }
    return Computed.makeComputed(Provide.provideToEffect(ref, provide), Provide.provideToStream(ref, provide), Provide.provideToEffect(ref.version, provide));
});
export const provideService = dual(3, function provideService(ref, tag, services) {
    return provide_(ref, Provide.ProvideService(tag, services));
});
export const provideServiceEffect = dual(3, function provideServiceEffect(ref, tag, service) {
    return provide_(ref, Provide.ProvideServiceEffect(tag, service));
});
class SimpleTransform extends VariancesImpl {
    ref;
    transformEffect;
    transformStream;
    shutdown;
    awaitShutdown;
    subscriberCount;
    version;
    changes;
    runUpdates;
    constructor(ref, transformEffect, transformStream) {
        super();
        this.ref = ref;
        this.transformEffect = transformEffect;
        this.transformStream = transformStream;
        this.shutdown = ref.shutdown;
        this.awaitShutdown = ref.awaitShutdown;
        this.subscriberCount = ref.subscriberCount;
        this.version = ref.version;
        this.changes = transformStream(ref.changes);
        this.runUpdates = ref.runUpdates;
    }
    toEffect() {
        return this.transformEffect(this.ref);
    }
    get get() {
        return this;
    }
}
export const withSpan = dual((args) => isSubscriptionRef(args[0]), (ref, name, options) => new SimpleTransform(ref, Effect.withSpan(name, options), Stream.withSpan(name, options)));
export const drop = dual(2, function drop(ref, n) {
    return isSubscriptionRef(ref)
        ? new SimpleTransform(ref, identity, Stream.drop(n))
        : Computed.makeComputed(ref.get, Stream.drop(ref.changes, n), ref.version);
});
export const take = dual(2, function take(ref, n) {
    return isSubscriptionRef(ref)
        ? new SimpleTransform(ref, identity, Stream.take(n))
        : Computed.makeComputed(ref.get, Stream.take(ref.changes, n), ref.version);
});
//# sourceMappingURL=LazyRef.js.map