import { Cause, Chunk, Context, Data, Deferred, Effect, Effectable, Equal, Equivalence, ExecutionStrategy, Exit, Fiber, MutableRef, Option, Predicate, PubSub, Queue, Record, Scope, Stream, } from 'effect';
import { pipeArguments } from 'effect/Pipeable';
import { constVoid } from 'effect/Function';
const toDeepEquals = (u) => {
    switch (typeof u) {
        case 'object': {
            if (Predicate.isNullable(u)) {
                return u;
            }
            else if (Equal.symbol in u) {
                return u;
            }
            else if (Array.isArray(u)) {
                return Data.tuple(u.map(toDeepEquals));
            }
            else if (u instanceof Set) {
                return Data.tuple(Array.from(u, toDeepEquals));
            }
            else if (u instanceof Map) {
                return Data.tuple(Array.from(u, ([k, v]) => Data.tuple([toDeepEquals(k), toDeepEquals(v)])));
            }
            else if (u instanceof URLSearchParams) {
                return Data.tuple(Array.from(u.keys()).map((key) => Data.tuple([key, toDeepEquals(u.getAll(key))])));
            }
            else if (Symbol.iterator in u) {
                return Data.tuple(Array.from(u, toDeepEquals));
            }
            else {
                return Data.struct(Record.map(u, toDeepEquals));
            }
        }
        default:
            return u;
    }
};
/**
 * @internal
 */
export const deepEquals = (a, b) => {
    // Attempt reference equality first for performance
    if (Object.is(a, b))
        return true;
    return Equal.equals(toDeepEquals(a), toDeepEquals(b));
};
export class EffectBase extends Effectable.StructuralClass {
    _effect;
    commit() {
        return (this._effect ??= this.toEffect());
    }
}
export class SubscriptionRefCore {
    initial;
    pubsub;
    runtime;
    scope;
    deferredRef;
    semaphore;
    constructor(initial, pubsub, runtime, scope, deferredRef, semaphore) {
        this.initial = initial;
        this.pubsub = pubsub;
        this.runtime = runtime;
        this.scope = scope;
        this.deferredRef = deferredRef;
        this.semaphore = semaphore;
    }
    _fiber = undefined;
}
export function makeCore(initial, options) {
    return Effect.runtime().pipe(Effect.bindTo('runtime'), Effect.let('executionStrategy', () => options?.executionStrategy ?? ExecutionStrategy.parallel), Effect.bind('scope', ({ executionStrategy, runtime }) => Scope.fork(Context.get(runtime.context, Scope.Scope), executionStrategy)), Effect.bind('id', () => Effect.fiberId), Effect.map(({ id, runtime, scope }) => unsafeMakeCore(initial, id, runtime, scope, options)), Effect.tap((core) => Scope.addFinalizer(core.scope, Effect.provide(core.pubsub.shutdown, core.runtime.context))));
}
export function unsafeMakeCore(initial, id, runtime, scope, options) {
    const pubsub = pubsubWithReplay();
    const core = new SubscriptionRefCore(initial, pubsub, runtime, scope, unsafeMakeDeferredRef(id, getExitEquivalence(options?.eq ?? deepEquals), pubsub.lastValue), Effect.unsafeMakeSemaphore(1));
    // Initialize the deferred ref with the initial value if it's already available
    matchEffectPrimitive(initial, {
        Success: (a) => core.deferredRef.done(Exit.succeed(a)),
        Failure: (cause) => core.deferredRef.done(Exit.failCause(cause)),
        Left: (e) => core.deferredRef.done(Exit.fail(e)),
        Right: (a) => core.deferredRef.done(Exit.succeed(a)),
        Some: (a) => core.deferredRef.done(Exit.succeed(a)),
        Sync: constVoid,
        None: constVoid,
        Otherwise: constVoid,
    });
    return core;
}
export function matchEffectPrimitive(effect, matchers) {
    const eff = effect;
    switch (eff._tag) {
        case 'Success':
            return matchers.Success(eff.value);
        case 'Failure':
            return matchers.Failure(eff.cause);
        case 'Sync':
            return matchers.Sync(eff.i0);
        case 'Left':
            return matchers.Left(eff.left);
        case 'Right':
            return matchers.Right(eff.right);
        case 'Some':
            return matchers.Some(eff.value);
        case 'None':
            return matchers.None(new Cause.NoSuchElementException());
        default:
            return matchers.Otherwise(effect);
    }
}
export const getExitEquivalence = (A) => Equivalence.make((a, b) => {
    if (a._tag === 'Failure') {
        return b._tag === 'Failure' && Equal.equals(a.cause, b.cause);
    }
    else {
        return b._tag === 'Success' && A(a.value, b.value);
    }
});
// Here to wrap the pubsub with a last value ref which can be shared with the DeferredRef
class PubsubWithReplay {
    pubsub;
    lastValue;
    [Queue.EnqueueTypeId] = {
        _In: (_) => _,
    };
    subscriberCount = MutableRef.make(0);
    constructor(pubsub, lastValue) {
        this.pubsub = pubsub;
        this.lastValue = lastValue;
    }
    publish(value) {
        return this.pubsub.publish(value);
    }
    publishAll(elements) {
        return this.pubsub.publishAll(elements);
    }
    subscribe = Effect.suspend(() => {
        MutableRef.update(this.subscriberCount, (count) => count + 1);
        return this.pubsub.subscribe.pipe(Effect.map((dequeue) => Option.match(MutableRef.get(this.lastValue), {
            onNone: () => dequeue,
            onSome: (previous) => dequeuePrepend(dequeue, previous),
        })), Effect.ensuring(Effect.sync(() => MutableRef.update(this.subscriberCount, (count) => count - 1))));
    });
    offer(value) {
        return this.pubsub.offer(value);
    }
    unsafeOffer(value) {
        return this.pubsub.unsafeOffer(value);
    }
    offerAll(elements) {
        return this.pubsub.offerAll(elements);
    }
    capacity() {
        return this.pubsub.capacity();
    }
    isActive() {
        return this.pubsub.isActive();
    }
    get size() {
        return this.pubsub.size;
    }
    unsafeSize() {
        return this.pubsub.unsafeSize();
    }
    get isFull() {
        return this.pubsub.isFull;
    }
    get isEmpty() {
        return this.pubsub.isEmpty;
    }
    get isShutdown() {
        return this.pubsub.isShutdown;
    }
    get shutdown() {
        return this.pubsub.shutdown;
    }
    get awaitShutdown() {
        return this.pubsub.awaitShutdown;
    }
    pipe() {
        // biome-ignore lint/style/noArguments: This is a pipeable
        return pipeArguments(this, arguments);
    }
}
class DequeueWithPrepend extends EffectBase {
    dequeue;
    previous;
    previousUtilized = false;
    [Queue.DequeueTypeId] = {
        _Out: (_) => _,
    };
    constructor(dequeue, previous) {
        super();
        this.dequeue = dequeue;
        this.previous = previous;
    }
    toEffect() {
        return this.take;
    }
    take = Effect.suspend(() => {
        if (this.previousUtilized) {
            return this.dequeue.take;
        }
        else {
            this.previousUtilized = true;
            return Effect.succeed(this.previous);
        }
    });
    takeAll = Effect.suspend(() => {
        if (this.previousUtilized) {
            return this.dequeue.takeAll;
        }
        else {
            this.previousUtilized = true;
            return this.dequeue.takeAll.pipe(Effect.map((a) => Chunk.prepend(a, this.previous)));
        }
    });
    takeUpTo(max) {
        return Effect.suspend(() => {
            if (this.previousUtilized) {
                return this.dequeue.takeUpTo(max);
            }
            else {
                this.previousUtilized = true;
                return this.dequeue
                    .takeUpTo(max - 1)
                    .pipe(Effect.map((a) => Chunk.prepend(a, this.previous)));
            }
        });
    }
    takeBetween(min, max) {
        return Effect.suspend(() => {
            if (this.previousUtilized) {
                return this.dequeue.takeBetween(min, max);
            }
            else {
                this.previousUtilized = true;
                return this.dequeue
                    .takeBetween(min - 1, max - 1)
                    .pipe(Effect.map((a) => Chunk.prepend(a, this.previous)));
            }
        });
    }
    capacity() {
        return this.dequeue.capacity();
    }
    isActive() {
        return this.dequeue.isActive();
    }
    get size() {
        return Effect.suspend(() => {
            if (this.previousUtilized) {
                return this.dequeue.size;
            }
            else {
                return this.dequeue.size.pipe(Effect.map((size) => size + 1));
            }
        });
    }
    unsafeSize() {
        if (this.previousUtilized) {
            return this.dequeue.unsafeSize();
        }
        else {
            return this.dequeue.unsafeSize().pipe(Option.map((size) => size + 1));
        }
    }
    get isFull() {
        return this.dequeue.isFull;
    }
    get isEmpty() {
        return Effect.suspend(() => {
            if (this.previousUtilized) {
                return this.dequeue.isEmpty;
            }
            else {
                return Effect.succeed(false);
            }
        });
    }
    get isShutdown() {
        return this.dequeue.isShutdown;
    }
    get shutdown() {
        return this.dequeue.shutdown;
    }
    get awaitShutdown() {
        return this.dequeue.awaitShutdown;
    }
    pipe() {
        // biome-ignore lint/style/noArguments: This is a pipeable
        return pipeArguments(this, arguments);
    }
}
export function dequeuePrepend(dequeue, previous) {
    return new DequeueWithPrepend(dequeue, previous);
}
export function pubsubWithReplay() {
    const lastValue = MutableRef.make(Option.none());
    const base = Effect.runSync(PubSub.unbounded());
    return new PubsubWithReplay(base, lastValue);
}
export function streamExit(stream) {
    return stream.pipe(Stream.map(Exit.succeed), Stream.catchAllCause((cause) => Stream.succeed(Exit.failCause(cause))));
}
export class DeferredRef extends EffectBase {
    id;
    eq;
    current;
    // Keep track of the latest value emitted by the stream
    version;
    deferred;
    constructor(id, eq, current) {
        super();
        this.id = id;
        this.eq = eq;
        this.current = current;
        this.reset();
    }
    toEffect() {
        return Effect.suspend(() => {
            const current = MutableRef.get(this.current);
            if (Option.isNone(current)) {
                return Deferred.await(this.deferred);
            }
            else {
                return current.value;
            }
        });
    }
    done(exit) {
        const current = MutableRef.get(this.current);
        MutableRef.set(this.current, Option.some(exit));
        if (Option.isSome(current) && this.eq(current.value, exit)) {
            return false;
        }
        Deferred.unsafeDone(this.deferred, exit);
        this.version += 1;
        return true;
    }
    reset() {
        MutableRef.set(this.current, Option.none());
        this.version = -1;
        if (this.deferred) {
            Deferred.unsafeDone(this.deferred, Exit.interrupt(this.id));
        }
        this.deferred = Deferred.unsafeMake(this.id);
    }
}
export function makeDeferredRef(eq) {
    return Effect.map(Effect.fiberId, (id) => new DeferredRef(id, eq, MutableRef.make(Option.none())));
}
export function unsafeMakeDeferredRef(id, eq, current) {
    return new DeferredRef(id, eq, current);
}
export class MulticastEffect extends EffectBase {
    effect;
    _fiber = null;
    _refCount = 0;
    constructor(effect) {
        super();
        this.effect = effect;
    }
    toEffect() {
        return Effect.suspend(() => {
            if (++this._refCount === 1) {
                return this.effect.pipe(Effect.forkDaemon, Effect.tap((fiber) => {
                    this._fiber = fiber;
                }), Effect.flatMap(Fiber.join), Effect.onExit(() => this.interrupt));
            }
            // biome-ignore lint/style/noNonNullAssertion: fiber is set by forkDaemon
            return Effect.fromFiber(this._fiber).pipe(Effect.onExit(() => this.interrupt));
        });
    }
    interrupt = Effect.suspend(() => {
        if (--this._refCount === 0 && this._fiber) {
            const interrupt = Fiber.interrupt(this._fiber);
            this._fiber = null;
            return interrupt;
        }
        return Effect.void;
    });
}
export function getSetDelete(core) {
    return {
        get: getOrInitializeCore(core, false),
        set: (a) => setCore(core, a),
        delete: deleteCore(core),
        version: Effect.sync(() => core.deferredRef.version),
    };
}
export function getOrInitializeCore(core, lockInitialize) {
    return Effect.suspend(() => {
        if (core._fiber === undefined && Option.isNone(MutableRef.get(core.deferredRef.current))) {
            return initializeCoreAndTap(core, lockInitialize);
        }
        else {
            return core.deferredRef;
        }
    });
}
function initializeCoreEffect(core, lock) {
    const initialize = Effect.onExit(Effect.provide(core.initial, core.runtime.context), (exit) => Effect.sync(() => {
        core._fiber = undefined;
        core.deferredRef.done(exit);
    }));
    return Effect.flatMap(Effect.forkIn(lock ? core.semaphore.withPermits(1)(initialize) : initialize, core.scope), (fiber) => Effect.sync(() => (core._fiber = fiber)));
}
function initializeCore(core, lock) {
    const onSuccess = (a) => {
        core.deferredRef.done(Exit.succeed(a));
        return Effect.succeed(Fiber.succeed(a));
    };
    const onCause = (cause) => {
        core.deferredRef.done(Exit.failCause(cause));
        return Effect.succeed(Fiber.failCause(cause));
    };
    const onError = (e) => onCause(Cause.fail(e));
    return matchEffectPrimitive(core.initial, {
        Success: onSuccess,
        Failure: onCause,
        Some: onSuccess,
        None: onError,
        Left: onError,
        Right: onSuccess,
        Sync: (f) => Effect.suspend(() => onSuccess(f())),
        Otherwise: () => initializeCoreEffect(core, lock),
    });
}
function initializeCoreAndTap(core, lock) {
    return Effect.zipRight(initializeCore(core, lock), tapEventCore(core, core.deferredRef));
}
function setCore(core, a) {
    const exit = Exit.succeed(a);
    return Effect.suspend(() => {
        if (core.deferredRef.done(exit)) {
            // If the value changed, send an event
            return Effect.as(sendEvent(core, exit), a);
        }
        else {
            // Otherwise, just return the current value
            return Effect.succeed(a);
        }
    });
}
export function interruptCore(core) {
    return Effect.fiberIdWith((id) => {
        core.deferredRef.reset();
        const closeScope = Effect.forkDaemon(Scope.close(core.scope, Exit.interrupt(id)));
        const interruptFiber = core._fiber ? Fiber.interruptFork(core._fiber) : Effect.void;
        const interruptSubject = core.pubsub.shutdown;
        return Effect.all([closeScope, interruptFiber, interruptSubject], {
            discard: true,
            concurrency: 'unbounded',
        });
    });
}
export function interruptCoreAndWait(core) {
    return Effect.fiberIdWith((id) => {
        core.deferredRef.reset();
        const closeScope = Scope.close(core.scope, Exit.interrupt(id));
        const interruptFiber = core._fiber ? Fiber.interrupt(core._fiber) : Effect.void;
        const interruptSubject = core.pubsub.awaitShutdown;
        return Effect.all([closeScope, interruptFiber, interruptSubject], {
            discard: true,
            concurrency: 'unbounded',
        });
    });
}
function deleteCore(core) {
    return Effect.suspend(() => {
        const current = MutableRef.get(core.deferredRef.current);
        core.deferredRef.reset();
        if (Option.isNone(current)) {
            return Effect.succeed(Option.none());
        }
        return Effect.sync(() => MutableRef.get(core.pubsub.subscriberCount)).pipe(Effect.flatMap((count) => count > 0 && !core._fiber ? initializeCore(core, false) : Effect.void), Effect.zipRight(Effect.asSome(current.value)));
    });
}
function tapEventCore(core, effect) {
    return effect.pipe(Effect.exit, Effect.tap((exit) => sendEvent(core, exit)), Effect.flatten);
}
export function sendEvent(core, exit) {
    return core.pubsub.publish(exit);
}
//# sourceMappingURL=internal.js.map